#include "ts_info.h"

#include "mpsys_parser.h"
#include "dvbsi_parser.h"

/* delivery names */
char * terrestrial_delivery_constellation_name (int c)
{
	switch (c)
	{
		case 0: return "qpsk";
		case 1: return "16";
		case 2: return "64";
	}
	return "reserved";
}
char * terrestrial_delivery_code_rate_name (int c)
{
	switch (c)
	{
		case 0: return "1_2";
		case 1: return "2_3";
		case 2: return "3_4";
		case 3: return "5_6";
		case 4: return "7_8";
	}
	return "reserved";
}

char satellite_delivery_constellation_name (int c)
{
	switch (c)
	{
		case 0: return 'H';
		case 1: return 'V';
		case 2: return 'L';
		case 3: return 'R';
	}
	return '?';
}

char *ca_system_name (int idin)
{
	int id = (idin >> 8);
	switch (id)
	{
		case 0x1: return "seca";
		case 0x5: return "viaccess";
		case 0x6: return "irdeto";
		case 0xb: return "conax";
		case 0x17: return "beta";
		case 0x18: return "nagra";
	}
	return 0;
}

int ca_has_id (conditional_access *ca)
{
	int id;
	assert (ca);
	id = ca->system_id >> 8;
	return (id == 1) || (id == 5);
}

int ca_id (conditional_access *ca)
{
	int id;
	assert (ca);
	assert (ca_has_id (ca));
	
	id = ca->system_id >> 8;
	if (id == 1)
	{
		if (buffer_iter_available_bytes (&ca->data)>2) 
		{
			buffer_iter b = ca->data;
			return buffer_iter_int_forth (&b, 16);
		}
	}
	if (id == 5)
	{
		buffer_iter b = ca->data;
		while (buffer_iter_available_bytes (&b) >= 2)
		{
			int type = buffer_iter_int_forth (&b, 8);
			int len = buffer_iter_int_forth (&b, 8);
			int avail = buffer_iter_available_bytes (&b);
			if (type == 0x14 && len >= 3 && avail >= 3)
			{
				return buffer_iter_int_forth (&b, 20) << 4;
			}
			buffer_iter_skip_bytes (&b, len < avail ? len : avail);
		}
	}
	return 0;
}

/* print info to XML */

void xml_print_descriptor (buffer *buf)
{
	if (service_parsable (buf))
	{
		service sdesc;
		service_parse (&sdesc, buf);

		fprintf (stdout, " <description provider_name='");
		buffer_iter_xml_text_dump (&sdesc.provider_name, stdout);
		fprintf (stdout, "' service_name='");
		buffer_iter_xml_text_dump (&sdesc.service_name, stdout);
		fprintf (stdout, "' type='%d'/>\n", sdesc.service_type);
	}
	else if (conditional_access_parsable (buf))
	{
		conditional_access ca;
		conditional_access_parse (&ca, buf);

		fprintf (stdout, " <ca_system_descriptor system_id='%d' ", ca.system_id);
		if (ca_system_name (ca.system_id))
		{
			fprintf (stdout, " system_name='%s'", ca_system_name (ca.system_id));
		}
		fprintf (stdout, " ecm_pid='%d'", ca.pid);
		if (ca_has_id (&ca))
		{
			fprintf (stdout, " id='%d'", ca_id (&ca));
		}
		fprintf (stdout, "/>");
	}
	else if (iso_language_parsable (buf))
	{
		iso_language lang;
		iso_language_parse (&lang, buf);
		fprintf (stdout, " <iso_639 language='%c%c%c' audio_type='%s' type='%d'/>\n",
			int_lowest (lang.language_code >> 16, 8),
			int_lowest (lang.language_code >> 8, 8),
			int_lowest (lang.language_code, 8),
			ts_program_descriptor_language_audio_type_name (lang.audio_type),
			lang.audio_type);
	}
}

void xml_print_descriptors (buffer_iter *iter_in)
{
	buffer_iter iter;
	for (iter = *iter_in;
			buffer_iter_available_bytes (&iter);
			ts_descriptor_forth (&iter))
	{
		buffer buf = buffer_iter_buffer_at (&iter); /* start of desc */
		xml_print_descriptor (&buf);
	}
}

void xml_print_network_satellite (satellite_delivery *del)
{
	fprintf (stdout, "<transponder type='S' freq='%d' polarity='%c' srate='%d'>\n", 
			del->frequency*10, 
			satellite_delivery_constellation_name (del->polarization),
			del->symbol_rate*100);
}

void xml_print_network_terrestrial (terrestrial_delivery *del)
{
		fprintf (stdout, "<transponder type='T' freq='%d' qam='%s' coderate='%s'>\n", 
			del->centre_frequency*10, 
			terrestrial_delivery_constellation_name (del->constellation),
			terrestrial_delivery_code_rate_name (del->code_rate_hp_stream));
}

void xml_print_network_cable (cable_delivery *del)
{
		fprintf (stdout, "<transponder type='C'>\n"); /* TODO: complete */
}


void xml_print_network_end (void)
{
	fprintf (stdout, "</transponder>\n");
}


void xml_print_service (struct service_description_services_st *se)
{
	assert (se);
	
	fprintf (stdout, "<service ca='%d' id='%d'>\n", se->free_ca_mode, se->service_id);
	xml_print_descriptors (&(se->descriptors));
}

void xml_print_service_end (void)
{
	fprintf (stdout, "</service>\n");
}

void xml_print_map (ts_program_map *pm)
{
	int i;
	
	xml_print_descriptors (&(pm->descriptors));
	for (i = 0; i < pm->streams_count; i++)
	{
		fprintf (stdout, " <stream type='%d' type_name='%s' pid='%d'>\n",
			pm->streams[i].stream_type,
			ts_stream_type_name (pm->streams[i].stream_type),
			pm->streams[i].es_pid);
		xml_print_descriptors (&(pm->streams[i].descriptors));
		fprintf (stdout, " </stream>\n");
	}
}

/* print concise info to text */

static void buffer_iter_padded_text_dump (buffer_iter *b, int minlen, FILE *f)
{
	int i = 0;
	buffer_iter iter = *b;
	while (buffer_iter_available_bytes (&iter))
	{
		unsigned char ch = (int) buffer_iter_int_forth (&iter, 8);
		fprintf (f, "%c", ch > 0x20 ? ch : ' '); 
		i ++;
	}
	while (i < minlen)
	{
		fprintf (f, " ");
		i ++;
	}
}

void short_text_print_network_satellite (satellite_delivery *del)
{
	fprintf (stdout, "-f %d -p %c -s %d\n", 
			del->frequency/100, 
			satellite_delivery_constellation_name (del->polarization),
			del->symbol_rate/10);
}

void short_text_print_network_terrestrial (terrestrial_delivery *del)
{
	fprintf (stdout, "-f %d -qam %s -cr %s\n", 
			del->centre_frequency/100, 
			terrestrial_delivery_constellation_name (del->constellation),
			terrestrial_delivery_code_rate_name (del->code_rate_hp_stream));
}

void short_text_print_network_cable (cable_delivery *del)
{
	fprintf (stdout, "cable\n");
}

void short_text_print_service (struct service_description_services_st *se)
{
	buffer_iter iter;

	assert (se);
	
	for (iter = se->descriptors;
			buffer_iter_available_bytes (&iter);
			ts_descriptor_forth (&iter))
	{
		buffer buf = buffer_iter_buffer_at (&iter); /* start of desc */
		if (service_parsable (&buf))
		{
			service sdesc;
			service_parse (&sdesc, &buf);

			buffer_iter_padded_text_dump (&sdesc.service_name, 20, stdout);
		}
	}
	fprintf (stdout, " %c ", se->free_ca_mode ? 'S':' ');
}

void short_text_print_end (void)
{
	fprintf (stdout, "\n");
}

void short_text_print_map (ts_program_map *pm)
{
	int i;	
	int has_video = 0;
	int has_audio = 0;
	
	/* video? */
	for (i = 0; i < pm->streams_count; i++)
	{
		if (pm->streams[i].stream_type == 2) 
		{ /* mpeg 2 video */
			fprintf (stdout, "TV %4d", pm->streams[i].es_pid);
			has_video++;
		}
		if ((pm->streams[i].stream_type == 3) || (pm->streams[i].stream_type == 4))
		{
			has_audio++;
		}
	}
	if (!has_video && has_audio)
	{
		fprintf (stdout, "RADIO -");
	}
	
	/* audio info */
	for (i = 0; i < pm->streams_count; i++)
	{
		if ((pm->streams[i].stream_type == 3) || (pm->streams[i].stream_type == 4))
		{ /* mpeg1, 2 audio */
			buffer_iter iter;
			fprintf (stdout, " %4d", pm->streams[i].es_pid);
			 		
			for (iter = pm->streams[i].descriptors;
					buffer_iter_available_bytes (&iter);
					ts_descriptor_forth (&iter))
			{
				buffer buf = buffer_iter_buffer_at (&iter); /* start of desc */
				if (iso_language_parsable (&buf))
				{
					iso_language lang;
					iso_language_parse (&lang, &buf);
					if (lang.audio_type == 0)
					{ /* normal audio */
						fprintf (stdout, " %c%c%c",
								int_lowest (lang.language_code >> 16, 8),
								int_lowest (lang.language_code >> 8, 8),
								int_lowest (lang.language_code, 8));
					}
				}
			}
		}
	}
}


void set_info_short_text (struct info_state_st *state)
{	
	state->print_network_satellite = short_text_print_network_satellite;
	state->print_network_terrestrial = short_text_print_network_terrestrial;
	state->print_network_cable = short_text_print_network_cable;
	state->print_network_end = short_text_print_end;
	state->print_service = short_text_print_service;
	state->print_service_end = short_text_print_end;
	state->print_map = short_text_print_map;
}

void set_info_xml (struct info_state_st *state)
{
	state->print_network_satellite = xml_print_network_satellite;
	state->print_network_terrestrial = xml_print_network_terrestrial;
	state->print_network_cable = xml_print_network_cable;
	state->print_network_end = xml_print_network_end;
	state->print_service = xml_print_service;
	state->print_service_end = xml_print_service_end;
	state->print_map = xml_print_map;
}

/* find out program map PID */
static int info_mapped (struct info_state_st *state, int id)
{
	int i;
	
	for (i = 0; i < state->programs_count; i++)
	{
		if (state->programs[i].program_number == id)
		{
			return state->programs[i].pid;
		}
	}
	return -1;
}

static void process_network_info (struct info_state_st *info_state, network_information *ni)
{
	int i;
	int found = 0;
	buffer_iter iter;
	
	assert (info_state);
	assert (ni);
	
	if (info_state->verbose)
	{
		fprintf (stderr, "NIT: section %d/%d\n", ni->section_number, ni->last_section_number);
	}
						
	/* find relevant transport info */
	for (i=0; i < ni->transports_count; i++)
	{			
		if (ni->transports[i].transport_stream_id == info_state->transport_stream_id)
		{
			for (iter = ni->transports[i].descriptors;
				buffer_iter_available_bytes (&iter);
				ts_descriptor_forth (&iter))
			{
				buffer buf = buffer_iter_buffer_at (&iter);
				if (satellite_delivery_parsable (&buf))
				{
					satellite_delivery del;
					satellite_delivery_parse (&del, &buf);
					(info_state->print_network_satellite) (&del);
					found++;
				}
				if (terrestrial_delivery_parsable (&buf))
				{
					terrestrial_delivery del;
					terrestrial_delivery_parse (&del, &buf);
					(info_state->print_network_terrestrial) (&del);
					found++;
				}
				if (cable_delivery_parsable (&buf))
				{
					cable_delivery del;
					cable_delivery_parse (&del, &buf);
					(info_state->print_network_cable) (&del);
					found++;
				}
			}
		}
	}
	
	/* other transponders */
	for (i=0; i < ni->transports_count; i++)
	{			
		for (iter = ni->transports[i].descriptors;
			buffer_iter_available_bytes (&iter);
			ts_descriptor_forth (&iter))
		{
			buffer buf = buffer_iter_buffer_at (&iter);
			if (satellite_delivery_parsable (&buf))
			{
				satellite_delivery del;
				satellite_delivery_parse (&del, &buf);
				(info_state->print_network_satellite) (&del);
				(info_state->print_network_end) ();
			}
			if (terrestrial_delivery_parsable (&buf))
			{
				terrestrial_delivery del;
				terrestrial_delivery_parse (&del, &buf);
				(info_state->print_network_terrestrial) (&del);
				(info_state->print_network_end) ();
			}
			if (cable_delivery_parsable (&buf))
			{
				cable_delivery del;
				cable_delivery_parse (&del, &buf);
				(info_state->print_network_cable) (&del);
				(info_state->print_network_end) ();
			}
		}
	}
	
	if(!found)
	{
		fprintf (stderr, "NIT: missing delivery info\n");
	}
}

static void process_program_association (struct info_state_st *info_state, ts_program_association *pat)
{
	int i;
	
	assert (info_state);
	assert (pat);
	
	if (info_state->verbose)
	{
		fprintf (stderr, "PAT: section %d/%d\n", pat->section_number, pat->last_section_number);
	}

	info_state->transport_stream_id = pat->transport_stream_id; 

	/* copy PATs from this section */
	for (i=0; (i < pat->programs_count) && (info_state->programs_count < TS_INFO_MAX_PROGRAMS); i++)
	{
		info_state->programs[info_state->programs_count] = pat->programs[i];
		(info_state->programs_count)++;
	}
}

/* section loop management */

static void info_section_reset (struct info_state_st *info_state)
{
	assert (info_state);
	info_state->section = -1;
	info_state->version = -1;
}

static void info_section_init (struct info_state_st *info_state, int is_current, int version_number)
{
	assert (info_state);
	assert (is_boolean (is_current));

	if (is_current && (info_state->version < 0))
	{
		info_state->section = 0;
		info_state->version = version_number;
	}
}

static int info_section_applies (struct info_state_st *info_state, int version, int section)
{
	assert (info_state);
	return (info_state->version == version) && (info_state->section == section);
}

/* state of service within current service description table */

static int info_service_has_next (struct info_state_st *info_state)
{
	int i;
	assert (info_state);
	
	for (i = info_state->services_next; i < info_state->service.services_count; i++)
	{
		return info_mapped (info_state, info_state->service.services[i].service_id) > 0;
	}
	return 0;
}

static void info_service_process (struct info_state_st *info_state)
{
	assert (info_state);
	assert (info_service_has_next (info_state));
	assert (info_state->status == TS_INFO_MAP);
	
	/* print end of previous service */
	if ((info_state->services_next > 0) || (info_state->section > 0))
	{
		(info_state->print_service_end)();
	}
	
	/* find and process next */
	info_state->service_info_id = 0;
	while (info_state->service_info_id == 0)
	{
		assert (info_state->services_next < info_state->service.services_count);
		info_state->service_info_id =
				info_mapped (info_state, info_state->service.services[info_state->services_next].service_id);
		if (info_state->service_info_id > 0)
		{
			(info_state->print_service)(&(info_state->service.services[info_state->services_next]));
		}
		(info_state->services_next)++;
	}

	/* 'next' */
	assert (info_state->service_info_id > 0);	
	assert (info_state->status == TS_INFO_MAP);
}

/* reset info */
void reset_info (struct info_state_st *result)
{
	result->status = TS_INFO_PAT;
	
	result->transport_stream_id = 0;
	result->programs_count = 0;	
	result->services_next = 0;
	result->service_info_id = 0;
	
	info_section_reset (result);
	
	/* default printer */
	result->do_nit = 0;
	result->verbose = 1;
	set_info_short_text (result);
}

/* current PID we're listening to */
int info_pid (struct info_state_st *info)
{
	switch (info->status)
	{
		case TS_INFO_PAT: return TS_PID_PAT;
		case TS_INFO_NIT: return TS_PID_NIT;
		case TS_INFO_SDT: return TS_PID_SDT;
		case TS_INFO_MAP: return info->service_info_id;
		case TS_INFO_FINISHED: return -1;
	}
	return -1;
}

/* Process a payload, taking into account only program information. 
   
   This is a state machine waiting for the next packet type. The 
   states are like:
   
   - wait for PAT (gets map of service id -> pid with stream info)
      then look for SDT
   - wait for SDT (gets services names and ids)
      then look for INFO for first service
   - wait for INFO for a given service
      then service = next_one, loop in this state until no services
   
   In PAT/SDT/NIT state we loop for each section (the information 
   may come in several packets which are numbered 0-max with max 
   known in each packet.
*/

void process_info (struct info_state_st *info_state, int payload_id, buffer *payload)
{
	buffer_iter payload_iter;
	buffer buf;
	
	assert (info_state);
	assert (payload);
	
	buffer_iter_reset (&payload_iter, payload);
	ts_pointer_skip (&payload_iter);

	buf = buffer_iter_buffer_at (&payload_iter);

	if ((info_state->status == TS_INFO_PAT) && (payload_id == TS_PID_PAT))
	{	
		/* collect service ID -> map PIDs table */
		if (ts_program_association_parsable (&buf))
		{
			ts_program_association pat; 
			ts_program_association_parse (&pat, &buf);

			info_section_init (info_state, pat.current, pat.version_number);				
			if (info_section_applies (info_state, pat.version_number, pat.section_number))
			{
				process_program_association (info_state, &pat);

				/* next section */
				(info_state->section)++;

				if (info_state->section > pat.last_section_number)
				{
					/* finished with all sections of PAT, next SDT */
					info_section_reset (info_state);
					info_state->status = info_state->do_nit ? TS_INFO_NIT : TS_INFO_SDT;
				}
			}
		}
		else
		{
			fprintf (stderr, "can't parse program association!\n");
		}
	}
	else if ((info_state->status == TS_INFO_NIT) && (payload_id == TS_PID_NIT))
	{
		if (network_information_parsable (&buf))
		{
			network_information ni;
			network_information_parse (&ni, &buf);

			if (ni.table_id == 0x40) /* this network */
			{
				info_section_init (info_state, ni.current_next_indicator, ni.version_number);
				if (info_section_applies (info_state, ni.version_number, ni.section_number))
				{
					process_network_info (info_state, &ni);

					(info_state->section)++;
					if (info_state->section > ni.last_section_number)
					{ /* next table */
						info_section_reset (info_state);
						info_state->status = TS_INFO_SDT;
					}
				}
			}
			else if (info_state->verbose)
			{
				fprintf (stderr, "NIT: section for other network\n");
			}
		}
		else
		{
			fprintf (stderr, "can't parse network info %s\n", 
					network_information_parsable_error (&buf));
		}
	}
	else if ((info_state->status == TS_INFO_SDT) && (payload_id == TS_PID_SDT))
	{	
		/* get a service to process from service table */
		if (service_description_parsable (&buf))
		{
			service_description sd;
			service_description_parse (&sd, &buf);

			if (sd.table_id == 0x42) /* current */
			{	
				info_section_init (info_state, sd.current_next_indicator, sd.version_number);
				if (info_section_applies (info_state, sd.version_number, sd.section_number))
				{
					if (info_state->verbose)
					{
						fprintf (stderr, "SDT: section %d/%d\n", sd.section_number, sd.last_section_number);
					}
					   
					/* save buffer in info_state and reparse service */
					{
						buffer sbuf;
						buffer_reset (&sbuf, info_state->service_buffer, sizeof (info_state->service_buffer));
						buffer_memory_copy (&sbuf, &buf);
						service_description_parse (&info_state->service, &sbuf);
					}
					info_state->services_next = 0;

					if (info_service_has_next (info_state))
					{
						info_state->status = TS_INFO_MAP;
						info_service_process (info_state);
					}
					else
					{
						fprintf (stderr, "SDT: section with no service?\n");
						info_state->section++;
						if (info_state->section > sd.last_section_number)
						{
							(info_state->print_network_end)();
							info_state->status = TS_INFO_FINISHED;
						}
					}
				}
			}
		}
		else
		{
			fprintf (stderr, "can't parse service description: %s!\n",
				service_description_parsable_error (&buf));
		}
	}
	else if ((info_state->status == TS_INFO_MAP) && (payload_id == info_state->service_info_id))
	{
		if (ts_program_map_parsable (&buf))
		{
			ts_program_map pm;
			ts_program_map_parse (&pm, &buf);

			(info_state->print_map)(&pm);

			if (info_service_has_next (info_state))
			{
				info_service_process (info_state);
				assert (info_state->status == TS_INFO_MAP); /* loop in MAP */	
			}
			else
			{
				/* next SDT section */
				info_state->status = TS_INFO_SDT;					
				(info_state->section)++;

				if (info_state->section > info_state->service.last_section_number)
				{
					(info_state->print_service_end)(); /* if there is >1? */
					if (info_state->do_nit)
					{
						(info_state->print_network_end)();
					}
					info_state->status = TS_INFO_FINISHED;
				}
			}
		}
		else
		{
			fprintf (stderr, "can't parse program map!\n");
		}
	}
}

